/*
 * TeamSpeak client Hello 3D example
 *
 * Copyright (c) 2007-2017 TeamSpeak Systems
 */

 #ifndef _WIN32
#include <unistd.h>
#endif
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include <teamspeak/public_definitions.h>
#include <teamspeak/clientlib_publicdefinitions.h>
#include <teamspeak/public_errors.h>
#include <teamspeak/clientlib.h>

#include "common.h"

/*A global variable that indicates the current connection status. Note that in
  a real world scenario, one would not use global variables.*/
int connection_status = STATUS_DISCONNECTED;

/*A global variable holding a pointer to the current TeamSpeak client identity*/
char* identity = NULL;

/*
* Callback for connection status change.
* Connection status switches through the states STATUS_DISCONNECTED, STATUS_CONNECTING, STATUS_CONNECTED and STATUS_CONNECTION_ESTABLISHED.
*
* Parameters:
*   connection_id             - Server connection handler ID
*   newStatus                 - New connection status, see the enum ConnectStatus in clientlib_publicdefinitions.h
*   errorNumber               - Error code. Should be zero when connecting or actively disconnection.
*                               Contains error state when losing connection.
*/
void onConnectStatusChangeEvent(uint64 connection_id, int new_status, unsigned int connection_error)
{
    printf("Connect status changed: %llu %d %u\n", (unsigned long long)connection_id, new_status, connection_error);

    if (new_status == STATUS_DISCONNECTED)
    {
        if (connection_error == ERROR_failed_connection_initialisation)
        {
            printf("Looks like there is no server running.\n");
        }
    }

    /*Update the global variable to the new state so the wait_for_connection
     function below can use it.*/
    connection_status = new_status;
}

void onTalkStatusChangeEvent(uint64 connection_id, int status, int is_received_whisper, anyID client_id)
{
    UNUSED(is_received_whisper);

    if (status == STATUS_TALKING)
    {
        anyID my_id;
        if (CHECK_ERROR(ts3client_getClientID(connection_id, &my_id)))
        {
            if (client_id != my_id)
            {
                TS3_VECTOR position;
                get_random_position_on_circle(&position);
                printf("Setting position to %f, %f, %f\n", position.x, position.y, position.z);
                CHECK_ERROR_FATAL(ts3client_channelset3DAttributes(connection_id, client_id, &position));
            }
        }
    }
}

void onCustom3dRolloffCalculationClientEvent(uint64 connection_id, anyID client_id, float distance, float* volume)
{
    UNUSED(connection_id);
    UNUSED(client_id);

    if (distance > 100)
        *volume = 0.0f;
    else if (*volume < 0.5f)
        *volume = 0.5f;
}

/* Initialize client lib with callbacks */
void ts_init(const char* soundbackends_path)
{
    /* Create struct for callback function pointers */
    struct ClientUIFunctions funcs;

    /* Initialize all callbacks with NULL */
    memset(&funcs, 0, sizeof(struct ClientUIFunctions));

    /* Callback function pointers */
    /* It is sufficient to only assign those callback functions you are using. When adding more callbacks, add those function pointers here. */
    funcs.onConnectStatusChangeEvent = onConnectStatusChangeEvent;
    funcs.onTalkStatusChangeEvent = onTalkStatusChangeEvent;

    CHECK_ERROR_FATAL(ts3client_initClientLib(&funcs, NULL, LogType_CONSOLE, NULL, soundbackends_path));

    /* Create a new client identity */
    /* In your real application you should do this only once, store the assigned identity locally and then reuse it. */
    CHECK_ERROR_FATAL(ts3client_createIdentity(&identity));
}

/*Create new server connection handler and connect it to the server, return connection id*/
uint64 ts_connect()
{
    uint64 connection_id = 0;
    /* Spawn a new server connection handler using the default port and store the server ID */
    CHECK_ERROR_FATAL(ts3client_spawnNewServerConnectionHandler(0, &connection_id));

    /* Connect to server on localhost:9987 with nickname "client", no default channel, no default channel password and server password "secret" */
    CHECK_ERROR_FATAL(ts3client_startConnection(connection_id, identity, "localhost", 9987, "client", NULL, "", "secret"));

    return connection_id;
}

/*A function to put the current thread into a wait loop until the ts3 connection
  has either fully connected, or there was a connection error (signaled by a 
  disconnect.
  returns 0 on error, 1 on connection fully established.
  */
bool wait_for_connection()
{
    /*Wait for connection_status to be "STATUS_CONNECTION_ESTABLISHED" which means fully connected */
    while (connection_status != STATUS_CONNECTION_ESTABLISHED)
    {
        /*If there was a connection error, the state will go to "STATUS_DISCONNECTED" so it is needed to check for that*/
        if (connection_status == STATUS_DISCONNECTED)
        {
            /*connection error occurred.*/
            return false;
        }
        sleep_ms(20);
    }

    return true;
}

/*Free used memory and shutdown the connection*/
void ts_shutdown()
{
    ts3client_freeMemory(identity);  /* Release dynamically allocated memory */
    identity = NULL;

    /* Shutdown client lib */
    CHECK_ERROR_FATAL(ts3client_destroyClientLib());
}

void start_audio(uint64 connection_id)
{
    /* Open default playback device */
    /* Passing empty string for mode and NULL or empty string for device will open the default device */
    CHECK_ERROR_FATAL(ts3client_openPlaybackDevice(connection_id, "", NULL));

    /* Open default capture device */
    /* Passing empty string for mode and NULL or empty string for device will open the default device */
    CHECK_ERROR_FATAL(ts3client_openCaptureDevice(connection_id, "", NULL));
}

void stop_audio(uint64 connection_id)
{
    // Close the default capture device
    CHECK_ERROR_FATAL(ts3client_closeCaptureDevice(connection_id));

    // Close the default playback device
    CHECK_ERROR_FATAL(ts3client_closePlaybackDevice(connection_id));
}

void set_listener_position(uint64 connection_id)
{
    TS3_VECTOR position;
    position.x = 0.0f;
    position.y = 1.0f;
    position.z = 0.0f;
    
    TS3_VECTOR forward;
    forward.x = 0.0f;
    forward.y = 1.0f;
    forward.z = 1.0f;

    TS3_VECTOR up;
    up.x = 0.0f;
    up.y = 1.5f;
    up.z = 0.0f;

    CHECK_ERROR_FATAL(ts3client_systemset3DListenerAttributes(connection_id, &position, &forward, &up));
}

int main(int argc, char **argv)
{
    UNUSED(argc);
    
    /* Resource path points to the SDK\bin directory to locate the soundbackends*/
    ts_init(program_path(argv[0]));

    create_circle_lookup_table(10);

    const uint64 connection_id = ts_connect();
    if (connection_id != 0)
    {
        start_audio(connection_id);

        if (wait_for_connection())
        {
            /* if we get here, we are fully connected */
            CHECK_ERROR_FATAL(ts3client_systemset3DSettings(connection_id, 1.0f, 1.0f));
            set_listener_position(connection_id);

            /* Wait for user input */
            printf("\n--- Press [q] to disconnect from server and exit ---\n");
            while (connection_status != STATUS_DISCONNECTED)
            {
                int c = getchar();
                switch (c)
                {
                    case 'q':
                    {
                        printf("\nDisconnecting from server...\n");
                        CHECK_ERROR(ts3client_stopConnection(connection_id, "leaving"));
                        break;
                    }
                }

                sleep_ms(20);
            }
        }

        stop_audio(connection_id);
        /* Destroy server connection handler */
        CHECK_ERROR_FATAL(ts3client_destroyServerConnectionHandler(connection_id));
    }

    ts_shutdown();
    return 0;
}
